"""
Easily composable and reactive UI utilities using RxPy and PyQt5.

This makes UI prototyping *MUCH* faster. In order to log IDs so that you can 
attach subscriptions after the fact, you will need to use the CollectUI
context manager.

An example is as follows, showing the currently available widgets. If you don't 
need to attach callbacks, you can get away without using the context manager.

```
ui = {}
with CollectUI(ui):
    test_widget = grid(
        group(
            text_edit('starting text', id='text'),
            line_edit('starting line', id='line'),
            combo_box(['A', 'B', 'C'], id='combo'),
            spin_box(5, id='spinbox'),
            radio_button('A Radio', id='radio'),
            check_box('Checkbox', id='check'),
            slider(id='slider'),
            file_dialog(id='file'),
            button('Send Text', id='submit')
        ),
        widget=self,
    )
```

"Forms" can effectively be built by building an observable out of the subjects in the 
UI. We have a `submit` function that makes creating such an observable simple.

```
submit('submit', ['check', 'slider', 'file'], ui).subscribe(lambda item: print(item))
```

With the line above, whenever the button with id='submit' is pressed, we will log a 
dictionary with the most recent values of the inputs {'check','slider','file'} as a 
dictionary with these keys. This allows building PyQt5 "forms" without effort.
"""

from enum import Enum
from typing import Type, Callable

import enum

import rx.operators as ops
import rx

import pyqtgraph as pg

from typing import Optional
from collections import namedtuple
from dataclasses import dataclass

import functools
from PyQt5.QtWidgets import (
    QGridLayout,
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QTabWidget,
    QSplitter,
    QGroupBox,
    QLabel,
)

from PyQt5 import QtCore, QtGui
from PyQt5.QtCore import Qt

from .widgets import *

__all__ = (
    "CollectUI",
    "CursorRegion",
    # layouts
    "layout",
    "grid",
    "vertical",
    "horizontal",
    "splitter",
    # widgets
    "group",
    "label",
    "tabs",
    "button",
    "check_box",
    "combo_box",
    "file_dialog",
    "line_edit",
    "radio_button",
    "slider",
    "spin_box",
    "text_edit",
    "numeric_input",
    # Observable tools
    "submit",
    # @dataclass utils
    "layout_dataclass",
    "bind_dataclass",
    "PRETTY_KEYS",
    "pretty_key_event",
    "Key",
    "KeyboardModifier",
    "KeyBinding",
)

Key = Qt.Key
KeyboardModifier = Qt.KeyboardModifier


@dataclass
class KeyBinding:
    """
    A keybinding for the application.

    Args:
        label: A human readable description of the keybinding.
        chord: A list of keys that can be pressed to trigger the handler.
        handler: The function to call when the keybinding is triggered.
    """

    label: str
    chord: list[Key]
    handler: Callable


CursorMode = namedtuple(
    "CursorMode", ("label", "chord", "handler", "supported_dimensions")
)

PRETTY_KEYS = {}
for key, value in vars(QtCore.Qt).items():
    if isinstance(value, QtCore.Qt.Key):
        PRETTY_KEYS[value] = key.partition("_")[2]


def pretty_key_event(event) -> list[str]:
    """Key Event -> list[str] in order to be able to prettily print keys.

    Args:
        event

    Returns:
        The key sequence as a human readable string.
    """
    key_sequence = []

    key_name = PRETTY_KEYS.get(event.key(), event.text())
    if key_name not in key_sequence:
        key_sequence.append(key_name)

    return key_sequence


ACTIVE_UI = None


def ui_builder(f):
    """Decorator synergistic with CollectUI to make widgets which register themselves automatically."""

    @functools.wraps(f)
    def wrapped_ui_builder(*args, id=None, **kwargs):
        global ACTIVE_UI
        if id is not None:
            try:
                id, ui = id
            except ValueError:
                id, ui = id, ACTIVE_UI

        ui_element = f(*args, **kwargs)

        if id:
            ui[id] = ui_element

        return ui_element

    return wrapped_ui_builder


class CollectUI:
    """Allows for collecing UI elements into a dictionary with IDs automatically.

    This makes it very easy to keep track of relevant widgets in a dynamically generated
    layout as they are just entries in a dict.
    """

    def __init__(self, target_ui=None):
        """We don't allow hierarchical UIs here, so ensure there's none active and make one."""
        global ACTIVE_UI
        assert ACTIVE_UI is None

        self.ui = {} if target_ui is None else target_ui
        ACTIVE_UI = self.ui

    def __enter__(self):
        """Pass my UI tree to the caller so they can write to it."""
        return self.ui

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Reset the active UI."""
        global ACTIVE_UI
        ACTIVE_UI = None


@ui_builder
def layout(*children, layout_cls=None, widget=None) -> QWidget:
    """A convenience method for constructing a layout and a parent widget."""
    if layout_cls is None:
        layout_cls = QGridLayout

    if widget is None:
        widget = QWidget()

    internal_layout = layout_cls()

    for child in children:
        internal_layout.addWidget(_wrap_text(child))

    widget.setLayout(internal_layout)

    return widget


grid = functools.partial(layout, layout_cls=QGridLayout)
vertical = functools.partial(layout, layout_cls=QVBoxLayout)
horizontal = functools.partial(layout, layout_cls=QHBoxLayout)


@ui_builder
def splitter(first, second, direction=Qt.Vertical, size=None) -> QWidget:
    """A convenience method for making a splitter."""
    split_widget = QSplitter(direction)

    split_widget.addWidget(first)
    split_widget.addWidget(second)

    if size is not None:
        split_widget.setSizes(size)

    return split_widget


splitter.Vertical = Qt.Vertical
splitter.Horizontal = Qt.Horizontal


@ui_builder
def group(*args, label=None, layout_cls=None) -> QWidget:
    """A convenience method for making a GroupBox container."""
    if args:
        if isinstance(args[0], str):
            label = args[0]
            args = args[1:]

    if layout_cls is None:
        layout_cls = QVBoxLayout

    groupbox = QGroupBox(label)

    layout = layout_cls()

    for arg in args:
        layout.addWidget(arg)

    groupbox.setLayout(layout)
    return groupbox


@ui_builder
def label(text, *args, **kwargs) -> QWidget:
    """A convenience method for making a text label."""
    return QLabel(text, *args, **kwargs)


@ui_builder
def tabs(*children) -> QWidget:
    """A convenience method for making a tabs control."""
    widget = QTabWidget()
    for name, child in children:
        widget.addTab(child, name)

    return widget


@ui_builder
def button(text, *args) -> QWidget:
    """A convenience method for making a Button."""
    return SubjectivePushButton(text, *args)


@ui_builder
def check_box(text, *args) -> QWidget:
    """A convenience method for making a checkbox."""
    return SubjectiveCheckBox(text, *args)


@ui_builder
def combo_box(items, *args, name=None) -> QWidget:
    """A convenience method for making a select/ComboBox."""
    widget = SubjectiveComboBox(*args)
    widget.addItems(items)

    if name is not None:
        widget.setObjectName(name)

    return widget


@ui_builder
def file_dialog(*args) -> QWidget:
    """A convenience method for making a button which opens a file dialog."""
    return SubjectiveFileDialog(*args)


@ui_builder
def line_edit(*args) -> QWidget:
    """A convenience method for making a single line text input."""
    return SubjectiveLineEdit(*args)


@ui_builder
def radio_button(text, *args) -> QWidget:
    """A convenience method for making a RadioButton."""
    return SubjectiveRadioButton(text, *args)


@ui_builder
def slider(minimum=0, maximum=10, interval=None, horizontal=True) -> QWidget:
    """A convenience method for making a Slider."""
    widget = SubjectiveSlider(orientation=Qt.Horizontal if horizontal else Qt.Vertical)
    widget.setMinimum(minimum)
    widget.setMaximum(maximum)

    if interval:
        widget.setTickInterval(interval)

    return widget


@ui_builder
def spin_box(minimum=0, maximum=10, step=1, adaptive=True, value=None) -> QWidget:
    """A convenience method for making a SpinBox."""
    widget = SubjectiveSpinBox()

    widget.setRange(minimum, maximum)

    if value is not None:
        widget.subject.on_next(value)

    if adaptive:
        widget.setStepType(SubjectiveSpinBox.AdaptiveDecimalStepType)
    else:
        widget.setSingleStep(step)

    return widget


@ui_builder
def text_edit(text="", *args) -> QWidget:
    """A convenience method for making multiline TextEdit."""
    return SubjectiveTextEdit(text, *args)


@ui_builder
def numeric_input(
    value=0, input_type: type = float, *args, validator_settings=None
) -> QWidget:
    """A numeric input with input validation."""
    validators = {
        int: QtGui.QIntValidator,
        float: QtGui.QDoubleValidator,
    }
    default_settings = {
        int: {
            "bottom": -1e6,
            "top": 1e6,
        },
        float: {
            "bottom": -1e6,
            "top": 1e6,
            "decimals": 3,
        },
    }

    if validator_settings is None:
        validator_settings = default_settings.get(input_type)

    widget = SubjectiveLineEdit(str(value), *args)
    widget.setValidator(
        validators.get(input_type, QtGui.QIntValidator)(**validator_settings)
    )

    return widget


def _wrap_text(str_or_widget):
    return label(str_or_widget) if isinstance(str_or_widget, str) else str_or_widget


def _unwrap_subject(subject_or_widget):
    try:
        return subject_or_widget.subject
    except AttributeError:
        return subject_or_widget


def submit(gate: str, keys: list[str], ui: dict[str, QWidget]) -> rx.Observable:
    """
    Builds an observable with provides the values of `keys` as a dictionary when `gate`
    changes. Essentially models form submission in HTML.
    """
    if isinstance(gate, str):
        gate = ui[gate]

    gate = _unwrap_subject(gate)
    items = [_unwrap_subject(ui[k]) for k in keys]

    combined = items[0].pipe(
        ops.combine_latest(*items[1:]), ops.map(lambda vs: dict(zip(keys, vs)))
    )

    return gate.pipe(
        ops.filter(lambda x: x), ops.with_latest_from(combined), ops.map(lambda x: x[1])
    )


def _try_unwrap_value(v):
    try:
        return v.value
    except AttributeError:
        return v


def enum_option_names(enum_cls: Type[enum.Enum]) -> list[str]:
    names = [x for x in dir(enum_cls) if "__" not in x]
    values = [_try_unwrap_value(getattr(enum_cls, n)) for n in names]

    return [x[0] for x in sorted(zip(names, values), key=lambda x: x[1])]


def enum_mapping(enum_cls: Type[enum.Enum], invert=False):
    options = enum_option_names(enum_cls)
    d = dict([[o, _try_unwrap_value(getattr(enum_cls, o))] for o in options])
    if invert:
        d = {v: k for k, v in d.items()}
    return d


def _layout_dataclass_field(dataclass_cls, field_name: str, prefix: str):
    id_for_field = f"{prefix}.{field_name}"

    field = dataclass_cls.__dataclass_fields__[field_name]
    if field.type in [
        int,
        float,
    ]:
        field_input = numeric_input(value=0, input_type=field.type, id=id_for_field)
    elif field.type == str:
        field_input = line_edit("", id=id_for_field)
    elif issubclass(field.type, enum.Enum):
        enum_options = enum_option_names(field.type)
        field_input = combo_box(enum_options, id=id_for_field)
    elif field.type == bool:
        field_input = check_box(field_name, id=id_for_field)
    else:
        raise Exception("Could not render field: {}".format(field))

    return group(
        field_name,
        field_input,
    )


def layout_dataclass(dataclass_cls, prefix: Optional[str] = None) -> QWidget:
    """Renders a dataclass instance to QtWidgets.

    See also `bind_dataclass` below to get one way data binding to the instance.

    Args:
        dataclass_cls
        prefix

    Returns:
        The widget containing the layout for the dataclass.
    """
    if prefix is None:
        prefix = dataclass_cls.__name__

    return vertical(
        *[
            _layout_dataclass_field(dataclass_cls, field_name, prefix)
            for field_name in dataclass_cls.__dataclass_fields__
        ]
    )


def bind_dataclass(dataclass_instance, prefix: str, ui: dict[str, QWidget]):
    """
    One-way data binding between a dataclass instance and a collection of widgets in the
    UI.

    Sets the current UI state to the value of the Python dataclass instance, and sets up
    subscriptions to value changes on the UI so that any future changes are propagated
    to the dataclass instance.

    Args:
        dataclass_instance: Instance to link
        prefix: Prefix for widget IDs in the UI
        ui: Collected UI elements
    """
    relevant_widgets = {
        k.split(prefix)[1]: v for k, v in ui.items() if k.startswith(prefix)
    }
    for field_name, field in dataclass_instance.__dataclass_fields__.items():
        translate_from_field, translate_to_field = {
            int: (lambda x: str(x), lambda x: int(x)),
            float: (lambda x: str(x), lambda x: float(x)),
        }.get(field.type, (lambda x: x, lambda x: x))

        if issubclass(field.type, Enum):
            forward_mapping = dict(
                sorted(enum_mapping(field.type).items(), key=lambda x: int(x[1]))
            )
            inverse_mapping = {v: k for k, v in forward_mapping.items()}

            def extract_field(v):
                try:
                    return v.value
                except AttributeError:
                    return v

            translate_to_field = lambda x: forward_mapping[x]
            translate_from_field = lambda x: inverse_mapping[extract_field(x)]

        current_value = translate_from_field(getattr(dataclass_instance, field_name))
        w = relevant_widgets[field_name]

        # write the current value to the UI
        w.subject.on_next(current_value)

        # close over the translation function
        def build_setter(translate, name):
            def setter(value):
                try:
                    value = translate(value)
                except ValueError:
                    return

                setattr(dataclass_instance, name, value)

            return setter

        w.subject.subscribe(build_setter(translate_to_field, field_name))


class CursorRegion(pg.LinearRegionItem):
    """
    A wide cursor to support an indication of the binning width in image marginals.
    """

    sigRegionChanged: QtCore.pyqtBoundSignal

    def __init__(self, *args, **kwargs):
        """Start with a width of one pixel."""
        super().__init__(*args, **kwargs)
        self._region_width = 1
        self.lines[1].setMovable(False)

    def set_width(self, value):
        """
        Adjusts the region by moving the right boundary to a distance `value` from the
        left.
        """
        self._region_width = value
        start, _ = self.getRegion()
        self.setRegion((start, start + value))

    def lineMoved(self, i):
        """
        Issues that the region for the cursor changed when one line on the boundary
        moves.
        """
        if self.blockLineSignal:
            return

        self.lines[1].setValue(self.lines[0].value() + self._region_width)
        self.prepareGeometryChange()
        self.sigRegionChanged.emit(self)

    # shouldn't be used, but here until FitTool is refactored
    def set_location(self, value: tuple[float, float]):
        """
        Sets the location of the cursor without issuing signals.

        Retains the width of the region so that you can just drag the wide cursor around.
        """
        old = self.blockLineSignal
        self.blockLineSignal = True
        self.lines[0].setValue(value[0])
        self.lines[1].setValue(value[1])
        self.blockLineSignal = old



"""Calculates momentum bounds from the angular coordinates.

Mostly these are used as common helper routines to the coordinate conversion code,
which is responsible for actually outputing the desired bounds.
"""

import numpy as np

from arpes.constants import K_INV_ANGSTROM
from arpes.typing import DataType

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import xarray as xr

__all__ = (
    "calculate_kp_kz_bounds",
    "calculate_kx_ky_bounds",
    "calculate_kp_bounds",
    "full_angles_to_k",
    "full_angles_to_k_approx",
)


def full_angles_to_k(
    kinetic_energy, phi, psi, alpha, beta, theta, chi, inner_potential
):
    """Converts from the full set of standard PyARPES angles to momentum.

    More details on angle to momentum conversion can be found at
    `the momentum conversion notes <https://arpes.readthedocs.io/momentum-conversion>`_.

    Args:
        kinetic_energy ([float]): [kinetic energy]
        phi ([float]): [angle along analyzer]
        psi ([float]): [analyzer deflector angle]
        alpha ([float]): [analyzer rotation angle]
        beta ([float]): [scan angle perpendicular to theta]
        theta ([float]): [goniometer azimuthal angle]
        chi ([float]): [sample azimuthal angle]
        inner_potential ([float]): [material inner potential in eV]

    Returns:
        [(float, float, float)]: [(kx, ky, kz)]
    """
    theta, beta, chi, psi = theta, beta, -chi, psi

    # use the full direct momentum conversion
    sin_alpha, cos_alpha = np.sin(alpha), np.cos(alpha)
    sin_beta, cos_beta = np.sin(beta), np.cos(beta)
    sin_theta, cos_theta = np.sin(theta), np.cos(theta)
    sin_chi, cos_chi = np.sin(chi), np.cos(chi)
    sin_phi, cos_phi = np.sin(phi), np.cos(phi)
    sin_psi, cos_psi = np.sin(psi), np.cos(psi)

    vx = cos_alpha * cos_psi * sin_phi - sin_alpha * sin_psi
    vy = sin_alpha * cos_psi * sin_phi + cos_alpha * sin_psi
    vz = cos_phi * cos_psi

    # perform theta rotation
    vrtheta_x = cos_theta * vx - sin_theta * vz
    vrtheta_y = vy
    vrtheta_z = sin_theta * vx + cos_theta * vz

    # perform beta rotation
    vrbeta_x = vrtheta_x
    vrbeta_y = cos_beta * vrtheta_y - sin_beta * vrtheta_z
    vrbeta_z = sin_beta * vrtheta_y + cos_beta * vrtheta_z

    # Perform chi rotation
    vrchi_x = cos_chi * vrbeta_x - sin_chi * vrbeta_y
    vrchi_y = sin_chi * vrbeta_x + cos_chi * vrbeta_y
    vrchi_z = vrbeta_z

    v_par_sq = vrchi_x**2 + vrchi_y**2

    """
    velocity -> momentum in each of parallel and perpendicular directions
    in the perpendicular case, we need the value of the cos^2(zeta) for the polar declination
    angle zeta in the sample (emission) frame. The total in plane velocity v_parallel is proportional
    to sin(zeta), so by the trig identity:

    1 = cos^2(zeta) + sin^2(zeta)

    we may substitute cos^2(zeta) for 1 - sin^2(zeta) which is 1 - (vrchi_x **2 + vrchi_y ** 2) above.
    """
    k_par = K_INV_ANGSTROM * np.sqrt(kinetic_energy)
    k_perp = K_INV_ANGSTROM * np.sqrt(kinetic_energy * (1 - v_par_sq) + inner_potential)

    return k_par * vrchi_x, k_par * vrchi_y, k_perp * vrchi_z


def euler_to_kx(kinetic_energy, phi, beta, theta=0, slit_is_vertical=False):
    """Calculates kx from the phi/beta Euler angles given the experimental geometry."""
    if slit_is_vertical:
        return K_INV_ANGSTROM * np.sqrt(kinetic_energy) * np.sin(beta) * np.cos(phi)
    else:
        return K_INV_ANGSTROM * np.sqrt(kinetic_energy) * np.sin(phi + theta)


def euler_to_ky(kinetic_energy, phi, beta, theta=0, slit_is_vertical=False):
    """Calculates ky from the phi/beta Euler angles given the experimental geometry."""
    if slit_is_vertical:
        return (
            K_INV_ANGSTROM
            * np.sqrt(kinetic_energy)
            * (np.cos(theta) * np.sin(phi) + np.cos(beta) * np.cos(phi) * np.sin(theta))
        )
    else:
        return (
            K_INV_ANGSTROM
            * np.sqrt(kinetic_energy)
            * (np.cos(phi + theta) * np.sin(beta),)
        )


def euler_to_kz(
    kinetic_energy, phi, beta, theta=0, inner_potential=10, slit_is_vertical=False
):
    """Calculates kz from the phi/beta Euler angles given the experimental geometry."""
    if slit_is_vertical:
        beta_term = -np.sin(theta) * np.sin(phi) + np.cos(theta) * np.cos(
            beta
        ) * np.cos(phi)

    else:
        beta_term = np.cos(phi + theta) * np.cos(beta)

    return K_INV_ANGSTROM * np.sqrt(kinetic_energy * beta_term**2 + inner_potential)


def spherical_to_kx(
    kinetic_energy: np.ndarray, theta: np.float64, phi: np.ndarray
) -> np.ndarray:
    """Calculates kx from the sample spherical (emission, not measurement) coordinates."""
    return K_INV_ANGSTROM * np.sqrt(kinetic_energy) * np.sin(theta) * np.cos(phi)


def spherical_to_ky(kinetic_energy, theta, phi):
    """Calculates ky from the sample spherical (emission, not measurement) coordinates."""
    return K_INV_ANGSTROM * np.sqrt(kinetic_energy) * np.sin(theta) * np.sin(phi)


def spherical_to_kz(
    kinetic_energy: np.ndarray, theta: np.float64, phi: np.ndarray, inner_V: np.float64
) -> np.ndarray:
    r"""Calculates the out of plane momentum from sample spherical (not measurement) coordinates.

    K_INV_ANGSTROM encodes that k_z = \frac{\sqrt{2 * m * E_kin * \cos^2\theta + V_0}}{\hbar}

    Args:
        kinetic_energy
        theta
        phi
        inner_V

    Returns:
        The out of plane momentum, kz.
    """
    return K_INV_ANGSTROM * np.sqrt(kinetic_energy * np.cos(theta) ** 2 + inner_V)


def calculate_kp_kz_bounds(data: DataType):
    """Calculates kp and kz bounds for angle-hv Fermi surfaces."""
    phi_offset = data.S.phi_offset
    phi_min = np.min(data.coords["phi"].values) - phi_offset
    phi_max = np.max(data.coords["phi"].values) - phi_offset

    binding_energy_min, binding_energy_max = np.min(data.coords["eV"].values), np.max(
        data.coords["eV"].values
    )
    hv_min, hv_max = np.min(data.coords["hv"].values), np.max(data.coords["hv"].values)

    wf = data.S.work_function
    kx_min = min(
        spherical_to_kx(hv_max - binding_energy_max - wf, phi_min, 0),
        spherical_to_kx(hv_min - binding_energy_max - wf, phi_min, 0),
    )
    kx_max = max(
        spherical_to_kx(hv_max - binding_energy_max - wf, phi_max, 0),
        spherical_to_kx(hv_min - binding_energy_max - wf, phi_max, 0),
    )

    angle_max = max(abs(phi_min), abs(phi_max))
    inner_V = data.S.inner_potential
    kz_min = spherical_to_kz(hv_min + binding_energy_min - wf, angle_max, 0, inner_V)
    kz_max = spherical_to_kz(hv_max + binding_energy_max - wf, 0, 0, inner_V)

    return (
        (round(kx_min, 2), round(kx_max, 2)),  # kp
        (round(kz_min, 2), round(kz_max, 2)),  # kz
    )


def calculate_kp_bounds(data: DataType, phi_offset: float, perpendicular_offset: float):
    """Calculates kp bounds for a single ARPES cut."""
    phi_coords = data.coords["phi"].values - phi_offset
    phi_bounds = np.min(phi_coords), np.max(phi_coords)
    kinetic_energy_sqrt = np.sqrt(
        data.S.hv - data.S.work_function + data.coords["eV"].values.max()
    )

    kp_bounds: np.ndarray = (
        K_INV_ANGSTROM
        * kinetic_energy_sqrt
        * np.cos(perpendicular_offset)
        * np.sin(phi_bounds)
    )

    return kp_bounds.round(2)


def calculate_kx_ky_bounds(
    data: DataType, phi_offset: float, scan_offset: float, scan_axis: str = "beta"
):
    """Calculate the kx and ky range for a dataset with a fixed photon energy.

    This is used to infer the gridding that should be used for a k-space conversion.
    Based on Jonathan Denlinger's old codes

    Args
    ----
    data : DataType
        Dataset that includes a key indicating the photon energy of the scan

    Returns
    -------
        ((kx_low, kx_high,), (ky_low, ky_high,))
    """
    kinetic_energy_sqrt = np.sqrt(
        data.S.hv - data.S.work_function + data.coords["eV"].values.max()
    )

    phi_coords: "xr.DataArray" = data.S.lookup_coord("phi") - phi_offset
    scan_coords: "xr.DataArray" = data.S.lookup_coord(scan_axis) - scan_offset
    phi_checkpoints = [phi_coords.min(), phi_coords.mean(), phi_coords.max()]
    scan_checkpoints = [scan_coords.min(), scan_coords.mean(), scan_coords.max()]

    potential_kx_bounds = []
    potential_ky_bounds = []
    for phi_checkpoint in phi_checkpoints:
        potential_kx_bounds.append(
            np.sin(phi_checkpoint) * kinetic_energy_sqrt * K_INV_ANGSTROM
        )
        for scan_checkpoint in scan_checkpoints:
            potential_ky_bounds.append(
                np.sin(scan_checkpoint)
                * np.cos(phi_checkpoint)
                * kinetic_energy_sqrt
                * K_INV_ANGSTROM
            )
    return (min(potential_kx_bounds), max(potential_kx_bounds)), (
        min(potential_ky_bounds),
        max(potential_ky_bounds),
    )



"""Wraps Qt widgets in ones which use rx for signaling, Conrad's personal preference."""
import os
from rx.subject import BehaviorSubject, Subject

from PyQt5.QtWidgets import (
    QPushButton,
    QCheckBox,
    QComboBox,
    QSpinBox,
    QTextEdit,
    QSlider,
    QLineEdit,
    QRadioButton,
    QWidget,
    QFileDialog,
    QHBoxLayout,
)

__all__ = (
    "SubjectivePushButton",
    "SubjectiveCheckBox",
    "SubjectiveComboBox",
    "SubjectiveFileDialog",
    "SubjectiveLineEdit",
    "SubjectiveRadioButton",
    "SubjectiveSlider",
    "SubjectiveSpinBox",
    "SubjectiveTextEdit",
)


class SubjectiveComboBox(QComboBox):
    """A QComboBox using rx instead of signals."""

    def __init__(self, *args, **kwargs):
        """Wrap signals in ``rx.BehaviorSubject``s."""
        super().__init__(*args, **kwargs)
        self.subject = BehaviorSubject(self.currentData())
        self.currentIndexChanged.connect(lambda: self.subject.on_next(self.currentText()))


class SubjectiveSpinBox(QSpinBox):
    """A QSpinBox using rx instead of signals."""

    def __init__(self, *args, **kwargs):
        """Wrap signals in ``rx.BehaviorSubject``s."""
        super().__init__(*args, **kwargs)
        self.subject = BehaviorSubject(self.value())
        self.valueChanged.connect(self.subject.on_next)
        self.subject.subscribe(self.update_ui)

    def update_ui(self, value):
        """Forwards value change to the UI."""
        self.setValue(value)


class SubjectiveTextEdit(QTextEdit):
    """A QTextEdit using rx instead of signals."""

    def __init__(self, *args):
        """Wrap signals in ``rx.BehaviorSubject``s."""
        super().__init__(*args)
        self.subject = BehaviorSubject(self.toPlainText())
        self.textChanged.connect(lambda: self.subject.on_next(self.toPlainText()))
        self.subject.subscribe(self.update_ui)

    def update_ui(self, value):
        """Forwards value change to the UI."""
        if self.toPlainText() != value:
            self.setPlainText(value)


class SubjectiveSlider(QSlider):
    """A QSlider using rx instead of signals."""

    def __init__(self, *args, **kwargs):
        """Wrap signals in ``rx.BehaviorSubject``s."""
        super().__init__(*args, **kwargs)
        self.subject = BehaviorSubject(self.value())
        self.valueChanged.connect(self.subject.on_next)
        self.subject.subscribe(self.update_ui)

    def update_ui(self, value):
        """Forwards value change to the UI."""
        self.setValue(value)


class SubjectiveLineEdit(QLineEdit):
    """A QLineEdit using rx instead of signals."""

    def __init__(self, *args):
        """Wrap signals in ``rx.BehaviorSubject``s."""
        super().__init__(*args)
        self.subject = BehaviorSubject(self.text())
        self.textChanged[str].connect(self.subject.on_next)
        self.subject.subscribe(self.update_ui)

    def update_ui(self, value):
        """Forwards value change to the UI."""
        if value != self.text():
            self.setText(value)


class SubjectiveRadioButton(QRadioButton):
    """A QRadioButton using rx instead of signals."""

    def __init__(self, *args):
        """Wrap signals in ``rx.BehaviorSubject``s."""
        super().__init__(*args)
        self.subject = BehaviorSubject(self.isChecked())
        self.toggled.connect(lambda: self.subject.on_next(self.isChecked()))
        self.subject.subscribe(self.update_ui)

    def update_ui(self, value):
        """Forwards value change to the UI."""
        self.setChecked(value)


class SubjectiveFileDialog(QWidget):
    """A file dialog implemented in Qt supporting single or multiple file selection."""

    def __init__(self, *args, single=True, dialog_root=None):
        """Sets up the file dialog widget.

        In addition to wrapping signals in BehaviorSubject as we do elsewhere,
        we need to get a reasonable place to open dialogs when they are requested.

        This can be configured with `dialog_root`, otherwise we will use `os.getcwd()`.
        """
        if dialog_root is None:
            dialog_root = os.getcwd()

        super().__init__(*args)

        self.dialog_root = dialog_root
        self.subject = BehaviorSubject(None)

        layout = QHBoxLayout()
        self.btn = SubjectivePushButton("Open")
        if single:
            self.btn.subject.subscribe(on_next=lambda _: self.get_file())
        else:
            self.btn.subject.subscribe(on_next=lambda _: self.get_files())

        layout.addWidget(self.btn)
        self.setLayout(layout)

    def get_file(self):
        """Opens a dialog allowing a single from the user."""
        filename = QFileDialog.getOpenFileName(self, "Open File", self.dialog_root)

        self.subject.on_next(filename[0])

    def get_files(self):
        """Opens a dialog allowing multiple selections from the user."""
        dialog = QFileDialog()
        dialog.setFileMode(QFileDialog.AnyFile)

        if dialog.exec_():
            filenames = dialog.selectedFiles()
            self.subject.on_next(filenames)


class SubjectivePushButton(QPushButton):
    """A QCheckBox using rx instead of signals."""

    def __init__(self, *args, **kwargs):
        """Wrap signals in ``rx.BehaviorSubject``s."""
        super().__init__(*args)
        self.subject = Subject()
        self.clicked.connect(lambda: self.subject.on_next(True))


class SubjectiveCheckBox(QCheckBox):
    """A QCheckBox using rx instead of signals."""

    def __init__(self, *args, **kwargs):
        """Wrap signals in ``rx.BehaviorSubject``s."""
        super().__init__(*args)
        self.subject = BehaviorSubject(self.checkState())
        self.stateChanged.connect(self.subject.on_next)
        self.subject.subscribe(self.update_ui)

    def update_ui(self, value):
        """Forwards value change to the UI."""
        self.setCheckState(value)



"""Utilities for estimating quantities of interest when using a laser for photoemission."""

import pint
from arpes.constants import ureg
from typing import Optional

__all__ = ("electrons_per_pulse", "electrons_per_pulse_mira")

mira_frequency = 54.3 / ureg.microsecond


def electrons_per_pulse(
    photocurrent: pint.Quantity,
    repetition_rate: Optional[pint.Quantity],
    division_ratio: int = 1,
) -> float:
    """Calculates the number of photoemitted electrons per pulse for pulsed lasers.

    Either the pulse_rate or the division_ratio and the base_repetition_rate should be
    specified.

    Args:
        photocurrent: The photocurrent in `pint` current units (i.e. amps).
        repetition_rate: The repetition rate for the laser, in `pint` frequency units.
        division_ratio: The division_ratio for a pulse-picked laser. Optionally modifies the
          repetition rate used for the calculation.

    Returns:
        The expectation of the number of electrons emitted per pulse of the laser.
    """
    repetition_rate /= division_ratio

    eles_per_attocoulomb = 6.2415091
    atto_coulombs = (photocurrent / repetition_rate).to("attocoulomb")

    return (atto_coulombs * eles_per_attocoulomb).magnitude


def electrons_per_pulse_mira(
    photocurrent: pint.Quantity, division_ratio: int = 1
) -> float:
    """Specific case of `electrons_per_pulse` for Mira oscillators."""
    return electrons_per_pulse(photocurrent, mira_frequency, division_ratio)



"""Utility functions used during data preparation and loading."""
from .axis_preparation import *
from .coord_preparation import *
from .hemisphere_preparation import *
from .tof_preparation import *



"""Implements data loading for the Beamline 7 (MAESTRO) ARPES experiments at ALS.

Common code is provided by a base class reflecting DAQ similarities between micro- and nanoARPES
at MAESTRO. This is subclassed for the individual experiments to handle some subtle differences
in how nanoARPES handles its spatial coordiantes (they are hierarchical) and in the spectrometers.
"""

from typing import override
import numpy as np
import xarray as xr

from arpes.endstations import (
    HDF5Endstation,
    HemisphericalEndstation,
    SynchrotronEndstation,
)

__all__ = (
    "MAESTROMicroARPESEndstation",
    #    "MAESTRONanoARPESEndstation"
)


class MAESTROARPESEndstationBase(
    SynchrotronEndstation, HemisphericalEndstation, HDF5Endstation
):
    """Common code for the MAESTRO ARPES endstations at the Advanced Light Source."""

    PRINCIPAL_NAME = None  # skip me
    ALIASES = None  # skip me
    ANALYZER_INFORMATION = None

    @override
    def postprocess_scan(self, data: xr.Dataset, scan_desc: dict = None):
        data.attrs.update(self.ANALYZER_INFORMATION)

        if "GRATING" in data.attrs:
            data.attrs["grating_lines_per_mm"] = {
                "G201b": 600,
            }.get(data.attrs["GRATING"])

        if "scan_x" in data.coords:
            # data.sel(scan_x=slice(None, None, -1)) #J: this might be a simpler way to do all below
            for data_var in data.data_vars:
                if "spectrum" in data_var:
                    data[data_var].values = np.flip(
                        data[data_var].values, axis=data[data_var].dims.index("scan_x")
                    )

        data = super().postprocess_scan(data, scan_desc)

        return data


class MAESTROMicroARPESEndstation(MAESTROARPESEndstationBase):
    """Implements data loading at the microARPES endstation of ALS's MAESTRO."""

    PRINCIPAL_NAME = "ALS-BL7"
    ALIASES = ["BL7", "BL7.0.2", "ALS-BL7.0.2", "MAESTRO"]

    ANALYZER_INFORMATION = {
        # "analyzer": "R4000",
        # "analyzer_name": "Scienta R4000",
        "parallel_deflectors": False,
        "perpendicular_deflectors": True,
        "analyzer_radius": None,
        "analyzer_type": "hemispherical",
    }

    RENAME_KEYS = {
        "X": "x",
        "Y": "y",
        "Z": "z",
        "Beta": "beta",
        "LMOTOR0": "x",
        "LMOTOR1": "y",
        "LMOTOR2": "z",
        "LMOTOR3": "theta",
        "LMOTOR4": "beta",
        "LMOTOR5": "chi",
        "LMOTOR6": "alpha",
        "LMOTOR9": "psi",
        "Scan X": "scan_x",
        "Scan Y": "scan_y",
        "Scan Z": "scan_z",
        "mono_eV": "hv",
        "SF_hv": "hv",
        "SS_hv": "hv",
        "Slit Defl": "psi",
        "S_Volts": "volts",
        # probably need something like an attribute list for extraction
        "SFRGN0": "fixed_region_name",
        "SFE_0": "daq_center_energy",
        "SFLNM0": "lens_mode_name",
        "SFPE_0": "pass_energy",
        "UNDHARM": "undulator_harmonic",
        "RINGCURR": "beam_current",
        "SFFR_0": "frames_per_slice",
        "SFBA_0": "phi_prebinning",
        "SFBE0": "eV_prebinning",
        "LWLVNM": "daq_type",
        "pixel": "phi",
    }

    ATTR_TRANSFORMS = {
        "Start_t": lambda l: {
            "time": " ".join(l.split(" ")[1:]).lower(),
            "date": l.split(" ")[0],
        },
        "SS_ESlitN": lambda l: {
            "slit_number": int(l.split(" ")[0]),
            "slit_shape": l.split(" ")[-1].lower(),
            "slit_width": float(l.split(" ")[2]),
        },
    }

    MERGE_ATTRS = {
        "mcp_voltage": None,
        "repetition_rate": 5e8,
        "undulator_type": "elliptically_polarized_undulator",
        "undulator_gap": None,
        "undulator_z": None,
        "undulator_polarization": None,
    }

    @override
    def postprocess_scan(self, data: xr.Dataset, scan_desc: dict = None):
        data = super().postprocess_scan(data, scan_desc)

        conversion_factor = {
            "Angular30": 0.181 / 4,
            "Angular14": 0.0963 / 4,
            "Angular7NF": 0.04264 / 4,
        }.get(data.attrs["SSlnm0"], 180 / np.pi)

        if "phi" in data.coords:
            data.coords["phi"] = (
                (data.coords["phi"] - data.coords["phi"].mean())
                * conversion_factor
                * (np.pi / 180)
            )

        data_vars = list(data.data_vars.keys())
        spectra_names = [data_var for data_var in data_vars if "Spectra" in data_var]
        max_dims = 0
        for spectrum_name in spectra_names:
            if n_dims := len(data[spectrum_name].dims) > max_dims:
                max_dims = n_dims
                data.attrs["spectrum_name"] = spectrum_name
        data = data.rename_vars({data.attrs["spectrum_name"]: "intensity"})
        data.attrs["spectrum_name"] = "intensity"

        return data



"""A control which can provide a range or a single value (i.e. a half open range with ends equal)."""
from functools import partial

from PyQt5 import QtWidgets


__all__ = ["RangeOrSingleValueWidget"]


class RangeOrSingleValueWidget(QtWidgets.QGroupBox):
    """A UI control letting you set a single value or a range.

    Used for modeling single cuts or multi-cut scans in the BZ tool.
    """

    def __init__(self, parent=None, root=None, coordinate_name=None, value=None):
        """Configures and initializes inner widgts.

        Inernally, we use a checkbox, spinbox, and slider to model the UI controls here,
        the checkbox determines which UI control is the "active" one.
        """
        super().__init__(title=coordinate_name, parent=parent)

        self.layout = QtWidgets.QGridLayout(self)

        self.label = QtWidgets.QLabel("Value: ")
        self.spinbox = QtWidgets.QSpinBox()
        self.slider = QtWidgets.QSlider()
        self.checkbox = QtWidgets.QCheckBox("Ranged")

        self.root = root

        self.layout.addWidget(self.label)
        self.layout.addWidget(self.spinbox)
        self.layout.addWidget(self.slider)
        self.layout.addWidget(self.toggle)

        self._prevent_change_events = False

        self.slider.valueChanged.connect(partial(self.value_changed, source=self.slider))
        self.spinbox.valueChanged.connect(partial(self.value_changed, source=self.spinbox))
        self.checkbox.valueChanged.connect(partial(self.mode_changed, source=self.checkbox))

        self.recompute()

    def mode_changed(self, event, source):
        """Unused, currently."""
        pass

    def value_changed(self, event, source):
        """Responds to changes in the internal value."""
        if self._prevent_change_events:
            return

        self._prevent_change_events = True
        self.slider.setValue(source.value())
        self.spinbox.setValue(source.value())
        self._prevent_change_events = False
        self.recompute()
        if self.root is not None:
            self.root.update_cut()

    def recompute(self):
        """Recompute UI representation from inner values."""
        value = self.spinbox.value()
        self.label.setText("Value: {:.3g}".format(value))

    def value(self):
        """The inner value."""
        return self.spinbox.value()



"""Scrollable interactive comparison between DataArrays."""

import numpy as np
import scipy.ndimage.interpolation

import colorcet as cc
import xarray as xr
from arpes.plotting.interactive_utils import BokehInteractiveTool
from arpes.typing import DataType
from arpes.utilities.funcutils import Debounce
from arpes.utilities.normalize_datatype import normalize_to_spectrum

__all__ = ["compare"]


class ComparisonTool(BokehInteractiveTool):
    auto_zero_nans = False
    auto_rebin = False
    other = None
    compared = None

    def __init__(self, other, **kwargs):
        super().__init__()

        self.load_settings(**kwargs)

        self.app_main_size = self.settings.get("app_main_size", 600)
        self.other = other

    def tool_handler(self, doc):
        from bokeh.layouts import row, column
        from bokeh.models import widgets
        from bokeh.models.mappers import LinearColorMapper
        from bokeh.plotting import figure

        default_palette = self.default_palette
        difference_palette = cc.coolwarm

        intensity_slider = widgets.Slider(
            title="Relative Intensity Scaling", start=0.5, end=1.5, step=0.005, value=1
        )

        self.app_context.update(
            {
                "A": self.arr,
                "B": self.other,
                "compared": self.compared,
                "plots": {},
                "figures": {},
                "widgets": {},
                "data_range": self.arr.G.range(),
                "color_maps": {},
            }
        )

        self.color_maps["main"] = LinearColorMapper(
            default_palette,
            low=np.min(self.arr.values),
            high=np.max(self.arr.values),
            nan_color="black",
        )

        figure_kwargs = {
            "tools": ["reset", "wheel_zoom"],
            "plot_width": self.app_main_size,
            "plot_height": self.app_main_size,
            "min_border": 10,
            "toolbar_location": "left",
            "x_range": self.data_range["x"],
            "y_range": self.data_range["y"],
            "x_axis_location": "below",
            "y_axis_location": "right",
        }

        self.figures["A"] = figure(title="Spectrum A", **figure_kwargs)
        self.figures["B"] = figure(title="Spectrum B", **figure_kwargs)
        self.figures["compared"] = figure(title="Comparison", **figure_kwargs)

        self.compared = self.arr - self.other
        diff_low, diff_high = np.min(self.arr.values), np.max(self.arr.values)
        diff_range = np.sqrt((abs(diff_low) + 1) * (abs(diff_high) + 1)) * 1.5
        self.color_maps["difference"] = LinearColorMapper(
            difference_palette, low=-diff_range, high=diff_range, nan_color="white"
        )

        self.plots["A"] = self.figures["A"].image(
            [self.arr.values],
            x=self.data_range["x"][0],
            y=self.data_range["y"][0],
            dw=self.data_range["x"][1] - self.data_range["x"][0],
            dh=self.data_range["y"][1] - self.data_range["y"][0],
            color_mapper=self.color_maps["main"],
        )

        self.plots["B"] = self.figures["B"].image(
            [self.other.values],
            x=self.data_range["x"][0],
            y=self.data_range["y"][0],
            dw=self.data_range["x"][1] - self.data_range["x"][0],
            dh=self.data_range["y"][1] - self.data_range["y"][0],
            color_mapper=self.color_maps["main"],
        )

        self.plots["compared"] = self.figures["compared"].image(
            [self.compared.values],
            x=self.data_range["x"][0],
            y=self.data_range["y"][0],
            dw=self.data_range["x"][1] - self.data_range["x"][0],
            dh=self.data_range["y"][1] - self.data_range["y"][0],
            color_mapper=self.color_maps["difference"],
        )

        x_axis_name = self.arr.dims[0]
        y_axis_name = self.arr.dims[1]

        stride = self.arr.G.stride()
        delta_x_axis = stride["x"]
        delta_y_axis = stride["y"]

        delta_x_slider = widgets.Slider(
            title="{} Shift".format(x_axis_name),
            start=-20 * delta_x_axis,
            step=delta_x_axis / 2,
            end=20 * delta_x_axis,
            value=0,
        )

        delta_y_slider = widgets.Slider(
            title="{} Shift".format(y_axis_name),
            start=-20 * delta_y_axis,
            step=delta_y_axis / 2,
            end=20 * delta_y_axis,
            value=0,
        )

        @Debounce(0.5)
        def update_summed_figure(attr, old, new):
            # we don't actually use the args because we need to pull all the data out
            shifted = (intensity_slider.value) * scipy.ndimage.interpolation.shift(
                self.other.values,
                [
                    delta_x_slider.value / delta_x_axis,
                    delta_y_slider.value / delta_y_axis,
                ],
                order=1,
                prefilter=False,
                cval=np.nan,
            )
            self.compared = self.arr - xr.DataArray(
                shifted, coords=self.arr.coords, dims=self.arr.dims
            )

            self.compared.attrs.update(**self.arr.attrs)
            try:
                del self.compared.attrs["id"]
            except KeyError:
                pass

            self.app_context["compared"] = self.compared
            self.plots["compared"].data_source.data = {"image": [self.compared.values]}

        layout = column(
            row(
                column(self.app_context["figures"]["A"]),
                column(self.app_context["figures"]["B"]),
            ),
            row(
                column(self.app_context["figures"]["compared"]),
                column(
                    intensity_slider,
                    delta_x_slider,
                    delta_y_slider,
                ),
            ),
        )

        update_summed_figure(None, None, None)

        delta_x_slider.on_change("value", update_summed_figure)
        delta_y_slider.on_change("value", update_summed_figure)
        intensity_slider.on_change("value", update_summed_figure)

        doc.add_root(layout)
        doc.title = "Comparison Tool"


def compare(A: DataType, B: DataType):
    """Opens an interactive tool with a scrollable diff between arrays."""
    A = normalize_to_spectrum(A)
    attrs = A.attrs
    B = normalize_to_spectrum(B)

    # normalize total intensity
    TOTAL_INTENSITY = 1000000
    A = A / (A.sum(A.dims) / TOTAL_INTENSITY)
    B = B / (B.sum(B.dims) / TOTAL_INTENSITY)
    A.attrs.update(**attrs)

    tool = ComparisonTool(other=B)
    return tool.make_tool(A)



"""Some general purpose analysis routines otherwise defying categorization."""

import itertools
from collections import defaultdict

import numpy as np

import arpes.constants
import arpes.models.band
import arpes.utilities
import arpes.utilities.math
import xarray as xr
from arpes.provenance import update_provenance
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum, lift_spectrum
from arpes.utilities.math import fermi_distribution

from .filters import gaussian_filter_arr

__all__ = (
    "normalize_by_fermi_distribution",
    "symmetrize_axis",
    "condense",
    "rebin",
    "fit_fermi_edge",
)


@update_provenance("Normalized by the 1/Fermi Dirac Distribution at sample temp")
def normalize_by_fermi_distribution(
    data: DataType,
    max_gain=None,
    rigid_shift=0,
    instrumental_broadening=None,
    total_broadening=None,
):
    """Normalizes a scan by 1/the fermi dirac distribution.

    You can control the maximum gain with ``clamp``, and whether
    the Fermi edge needs to be shifted (this is for those desperate situations where you want something that
    "just works") via ``rigid_shift``.

    Args:
        data: Input
        max_gain: Maximum value for the gain. By default the value used
            is the mean of the spectrum.
        rigid_shift: How much to shift the spectrum chemical potential.
        instrumental_broadening: Instrumental broadening to use for
            convolving the distribution
    Pass the nominal value for the chemical potential in the scan. I.e. if the chemical potential is at BE=0.1, pass
    rigid_shift=0.1.

    Returns:
        Normalized DataArray
    """
    data = normalize_to_spectrum(data)

    if total_broadening:
        distrib = fermi_distribution(
            data.coords["eV"].values - rigid_shift,
            total_broadening / arpes.constants.K_BOLTZMANN_EV_KELVIN,
        )
    else:
        distrib = fermi_distribution(
            data.coords["eV"].values - rigid_shift, data.S.temp
        )

    # don't boost by more than 90th percentile of input, by default
    if max_gain is None:
        max_gain = min(np.mean(data.values), np.percentile(data.values, 10))

    distrib[distrib < 1 / max_gain] = 1 / max_gain
    distrib_arr = xr.DataArray(distrib, {"eV": data.coords["eV"].values}, ["eV"])

    if instrumental_broadening is not None:
        distrib_arr = gaussian_filter_arr(
            distrib_arr, sigma={"eV": instrumental_broadening}
        )

    return data / distrib_arr


@update_provenance("Symmetrize about axis")
def symmetrize_axis(data, axis_name, flip_axes=None, shift_axis=True):
    """Symmetrizes data across an axis.

    It would be better ultimately to be able
    to implement an arbitrary symmetry (such as a mirror or rotational symmetry
    about a line or point) and to symmetrize data by that method.

    Args:
        data
        axis_name
        flip_axes
        shift_axis

    Returns:
        Data after symmetrization procedure.
    """
    data = data.copy(deep=True)  # slow but make sure we don't bork axis on original
    data.coords[axis_name].values = (
        data.coords[axis_name].values - data.coords[axis_name].values[0]
    )

    selector = {}
    selector[axis_name] = slice(None, None, -1)
    rev = data.sel(**selector).copy()

    rev.coords[axis_name].values = -rev.coords[axis_name].values

    if flip_axes is None:
        flip_axes = []

    for axis in flip_axes:
        selector = {}
        selector[axis] = slice(None, None, -1)
        rev = rev.sel(**selector)
        rev.coords[axis].values = -rev.coords[axis].values

    return rev.combine_first(data)


@update_provenance("Condensed array")
def condense(data: xr.DataArray):
    """Clips the data so that only regions where there is substantial weight are included.

    In practice this usually means selecting along the ``eV`` axis, although other selections
    might be made.

    Args:
        data: xarray.DataArray

    Returns:
        The clipped data.
    """
    if "eV" in data.dims:
        data = data.sel(eV=slice(None, 0.05))

    return data


# J: TODO - rebin coordinates instead of simply taking the nth element
@update_provenance("Rebinned array")
def rebin(
    data: DataType,
    shape: dict | None = None,
    reduction: int | dict | None = None,
    interpolate=False,
    **kwargs
):
    """Rebins the data onto a different (smaller) shape.

    By default the behavior is to
    split the data into chunks that are integrated over. An interpolation option is also
    available.

    Exactly one of ``shape`` or ``reduction`` should be supplied.

    Dimensions corresponding to missing entries in ``shape`` or ``reduction`` will not
    be changed.

    Args:
        data
        interpolate: Use interpolation instead of integration
        shape: Target shape
        reduction: Factor to reduce each dimension by

    Returns:
        The rebinned data.
    """
    if isinstance(data, xr.Dataset):
        new_vars = {
            datavar: rebin(
                data[datavar],
                shape=shape,
                reduction=reduction,
                interpolate=interpolate,
                **kwargs
            )
            for datavar in data.data_vars
        }
        new_coords = {}

        for var in new_vars.values():
            new_coords.update(var.coords)
        return xr.Dataset(data_vars=new_vars, coords=new_coords, attrs=data.attrs)

    data = arpes.utilities.normalize_to_spectrum(data)

    if any(d in kwargs for d in data.dims):
        reduction = kwargs

    if interpolate:
        raise NotImplementedError("The interpolation option has not been implemented")

    assert shape is None or reduction is None

    if isinstance(reduction, int):
        reduction = {d: reduction for d in data.dims}

    if reduction is None:
        reduction = {}

    # we standardize by computing reduction from shape is shape was supplied.
    if shape is not None:
        reduction = {k: len(data.coords[k]) // v for k, v in shape.items()}

    # since we are not interpolating, we need to clip each dimension so that the reduction
    # factor evenly divides the real shape of the input data.
    slices = defaultdict(lambda: slice(None))

    if not data.dims:
        return data

    for dim, reduction_factor in reduction.items():
        remainder = len(data.coords[dim]) % reduction_factor
        if remainder != 0:
            slices[dim] = slice(None, -remainder)

    trimmed_data = data.data[tuple(slices[d] for d in data.dims)]
    trimmed_coords = {d: coord[slices[d]] for d, coord in data.indexes.items()}

    temp_shape = [
        [trimmed_data.shape[i] // reduction.get(d, 1), reduction.get(d, 1)]
        for i, d in enumerate(data.dims)
    ]
    temp_shape = itertools.chain(*temp_shape)
    reduced_data = trimmed_data.reshape(*temp_shape)

    for i in range(len(data.dims)):
        reduced_data = reduced_data.mean(i + 1)

    reduced_coords = {
        d: coord[:: reduction.get(d, 1)] for d, coord in trimmed_coords.items()
    }
    reduced_coords.update(
        {c: data.coords[c] for c in data.coords.keys() if c not in trimmed_coords}
    )

    return xr.DataArray(reduced_data, reduced_coords, data.dims, attrs=data.attrs)


@lift_spectrum
def normalize(spectrum: xr.DataArray) -> xr.DataArray:
    """Normalize spectrum so its minimum value is 0 and maximum value is 1."""
    return (spectrum - spectrum.min()) / (spectrum.max() - spectrum.min())



"""Image reading methods with different library fallbacks."""
import warnings

import numpy as np
import xarray as xr

__all__ = (
    "imread_to_xarray",
    "imread",
)


def imread(str_or_path) -> np.ndarray:
    """A wrapper around `opencv.imread` and `imageio.imread`.

    As compared to those, this method

    1. Falls back to the first available option on the system
    2. Normalizes OpenCV images to RGB format
    3. Removes the alpha channel from imageio.imread data

    Args:
        str_or_path: pathlib.Path or str containing the image to be read

    Returns:
        The image as an np.ndarray.
    """
    try:
        import cv2

        using_cv2 = True
    except ImportError:
        try:
            import imageio

            using_cv2 = False
        except ImportError as e:
            warnings.warn("You need OpenCV or imageio in order to read images in PyARPES.")
            raise e

    if using_cv2:
        arr = cv2.imread(str(str_or_path))
        return np.stack([arr[:, :, 2], arr[:, :, 1], arr[:, :, 0]], axis=-1)
    else:
        arr = imageio.imread(str(str_or_path))
        return arr[:, :, :3]


def imread_to_xarray(str_or_path) -> xr.DataArray:
    """Like `imread`, except that this function wraps the result into a xr.DataArray instance.

    The read instance has x (pixel), y (pixel), and color (['R', 'G', 'B']) dimensions.

    Args:
        str_or_path

    Returns:
        The data read to an `xr.DataArray` instance.
    """
    raw_arr = imread(str_or_path)
    sx, sy, _ = raw_arr.shape

    return xr.DataArray(
        raw_arr,
        coords={"x": np.asarray(range(sx)), "y": np.asarray(range(sy)), "color": ["r", "g", "b"]},
        dims=["x", "y", "color"],
    )
